#!/bin/bash
###############################################################################--c01
# This library "custom-additions-01-limit-system-partition-resize" helps
# to change the default behaviour of the Raspberry Pi OS and to limit
# root partition resize during installation by desired value.
# This library is an optional part of the "prepare-system-image-for-read-only-mode" library.
#
# This library is supplementary material to the article
# "How to configure Raspberry Pi OS to run in read-only mode".
# The article and this library can be found at https://altomaxtech.com/.
#
# More "custom-additions-*" to come.
# Please visit https://altomaxtech.com/ often to check out the news.
#
# Copyright (C) 2021, Vladimir Zlatov, Evgueni Tzvetanov, USA
# Email us at support@altomaxtech.com.
#
# This library is free software; you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General
# Public License along with this library;
# if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
###############################################################################
###############################################################################
# Notes. 
#    This library is able to automatically find Raspberry Pi OS zip files 
#    with images using observed current naming scheme (for example, like 
#    "2021-05-07-raspios-buster-armhf.zip" for Raspberry Pi OS with Desktop).
#    There is no guarantee that this scheme will be used in future releases 
#    of Raspberry Pi OS Team. 
#    The naming scheme is defined by get_default_rpi_os_filename_regex function.
#    The Raspberry Pi OS images can be downloaded from the developer's website:
#    https://www.raspberrypi.org/software/operating-systems/
#    This Library is suitable for any of the packages available there:
#       - Raspberry Pi OS with desktop and recommended software; 
#       - Raspberry Pi OS with desktop;
#       - Raspberry Pi OS Lite.
######################################################################
#    This library uses the /tmp directory for all intermediate files and mounts.
#
###############################################################################
# Known issues:
#   - In case of unexpected errors the script may terminate without proper
#     unmounting of currently mounted partition(s) of the image.
#     The incompleted image and temporary mount directories can be dismissed
#     assuming the system automatic cleaning care of temporary locations.
#     Or, the image and the directories can be removed manually if needed.
#   - The error handling is not consistent enough in some operators which may
#     allow next command(s) to proceed as normal without noticing that
#     the previous command has failed.
#     It is always a good idea to check out reports for errors.
#
#
###############################################################################

source "prepare-system-image-for-read-only-mode-functions-00.sh"


###############################################################################
modify_init_resize_sh() {
    # $1 - source full path
    # $2 - output full path (can be the same as the source's)
    # NOTE. The path of the file in the original image is "/usr/lib/raspi-config/init_resize.sh"

    local timestamp=$(date -u);
    local result=$(sed  "s/\s*TARGET_END=/### REMOVED at $timestamp ###&/1" $1);
    error_exit ${LINENO} "Failed to edit $1";
    result=$(echo "$result" | sed  '/\s*ROOT_PART_END=/a\  ROOT_PART_SIZE=$(echo "$ROOT_PART_LINE" | cut -d ":" -f 4)       ### ADDED ###\n  TARGET_END=$((ROOT_PART_SIZE * 2))       ### ADDED TO REPLACE ORIGINAL ###');
    echo "$result" > $2;
    error_exit ${LINENO} "Failed to save results to $2";
}


###############################################################################
modify_init_resize_sh_in_place() {
    # $1 - file system root mount point (dir without closing slash)
    # NOTE. The path of the file in the original image is "/usr/lib/raspi-config/init_resize.sh"

    local output_file_full_path="$1/usr/lib/raspi-config/init_resize.sh";
    error_exit ${LINENO} "Failed to initialize 'output_file_full_path' local variable.";
    modify_init_resize_sh "$output_file_full_path" "$output_file_full_path"
    error_exit ${LINENO} "Failed to modify_init_resize_sh for $output_file_full_path";
    sed -i "/^#!\/bin\/.*\$/a\
#\n\
###############################################################################\n\
# This script was modified automatically by\n\
# $(basename \"$0\")\n\
# at $(date -u).\n\
# See additional notes below on modified lines.\n\
###############################################################################\n" $output_file_full_path
    error_exit ${LINENO} "Failed to modify_init_resize_sh for $output_file_full_path";

}


###############################################################################
###############################################################################
main_additions_01() {

    local this_module_name="custom-additions-01-limit-system-partition-resize";
    local system_image_directory="/tmp";

    #################################################################################
    #
    echo "'$this_module_name' STARTED  at $(date)"

    exit_if_not_root;


    # Discover all available img files in the tmp_image_dir, then process them one by one.  
    local img_files_full_path_list=$(find_rpi_os_file_list  "$system_image_directory" ".img");


    if [ -z "$img_files_full_path_list" ]
    then
        terminate "No RPi OS image files have been discovered in the directory '$system_image_directory'";
    else {


        while IFS= read -r img_file_full_path; do {
            echo ""
            echo "================================================================================"
            echo "Processing image file '$img_file_full_path' ..."
            echo "===================================="
            echo ""

            check_if_img_file_exists_and_exit_if_not "$img_file_full_path";

            local tmp_mount_root_dir=$(mktemp -d -t rpi-mount-root---XXXXXXXXXXXX);

            if mount_original_image_to_tmp_root_dir "$img_file_full_path" "$tmp_mount_root_dir"; then {

                echo "Prepare for root partition resizing on the first boot..."
                modify_init_resize_sh_in_place "$tmp_mount_root_dir"

                umount "$tmp_mount_root_dir"
                echo "Unmount modified image Root Partition."
            }
            else
                echo "Mount FAILED for image Root Partition. ErrorCode=$?."
            fi

            echo "Removing tmp directories"
            rm -rf "$tmp_mount_root_dir"

            
        }
        done <<< "$img_files_full_path_list"

    }
    fi
            
    #
    #################################################################################

    echo "'$this_module_name' FINISHED at $(date)";
}

#################################################################################
#################################################################################
#################################################################################

main_additions_01;
