#!/bin/bash
###############################################################################
# This library "prepare-system-image-for-read-only-mode" helps to configure 
# an image of Raspberry Pi OS or other compatible Linux distributions 
# to run in read-only mode.
#
# This library is supplementary material to the article 
# "How to configure Raspberry Pi OS to run in read-only mode". 
# The article and this library can be found at https://altomaxtech.com/. 
#
# Copyright (C) 2021, Vladimir Zlatov, Evgueni Tzvetanov, USA 
# Email us at support@altomaxtech.com. 
# 
# This library is free software; you can redistribute it 
# and/or modify it under the terms of the GNU Lesser General Public 
# License as published by the Free Software Foundation; either 
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General 
# Public License along with this library; 
# if not, write to the Free Software Foundation, Inc., 
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
###############################################################################
###############################################################################
# Notes. 
#    This library is able to automatically find Raspberry Pi OS zip files 
#    with images using observed current naming scheme (for example, like 
#    "2021-05-07-raspios-buster-armhf.zip" for Raspberry Pi OS with Desktop).
#    There is no guarantee that this scheme will be used in future releases 
#    of Raspberry Pi OS Team. 
#    The naming scheme is defined by get_default_rpi_os_filename_regex function.
#    The Raspberry Pi OS images can be downloaded from the developer's website:
#    https://www.raspberrypi.org/software/operating-systems/
#    This Library is suitable for any of the packages available there:
#       - Raspberry Pi OS with desktop and recommended software; 
#       - Raspberry Pi OS with desktop;
#       - Raspberry Pi OS Lite.
######################################################################
#    This library uses the /tmp directory for all intermediate files and mounts.
#
###############################################################################
# Known issues:
#   - In case of unexpected errors the script may terminate without proper  
#     unmounting of currently mounted partition(s) of the image.
#     The incompleted image and temporary mount directories can be dismissed 
#     assuming the system automatic cleaning care of temporary locations.
#     Or, the image and the directories can be removed manually if needed.
#   - The error handling is not consistent enough in some operators which may 
#     allow next command(s) to proceed as normal without noticing that 
#     the previous command has failed.
#     It is always a good idea to check out reports for errors.
#      
# 
###############################################################################



###############################################################################
get_default_rpi_os_filename_regex() {
    
    echo "202[1-9]-\(0[1-9]\|1[0-2]\)-\([0-2][0-9]\|3[0-1]\)-raspios-.*-armhf.*";
}

###############################################################################
find_rpi_os_file_list() {
    # $1 - working directory (without closing slash)
    # $2 - file extention (".zip", ".img", etc...)
    
    local rpi_os_files=$(find "$1" -maxdepth 1 -regex ".*\($get_default_rpi_os_filename_regex\)"+"$2");
    echo "$rpi_os_files";
}

###############################################################################
get_image_partitions_count() {
    # $1 - full path of an image file (<filename>.img)

    local result=$(partx -s -o NR,TYPE -g "$1");
    local count=0;
    while IFS= read -r line; do
        ((count++));
        #echo "... $line ..."
    done <<< "$result"
    echo "$count";
}

###############################################################################
get_image_partition_type() {
    # $1 - full path of an image file (<filename>.img)
    # $2 - partition number. Expecting: 1 - /boot (vfat), or 2 - /root (ext4)

    local result=$(partx -s -o TYPE -g -n "$2:$2"  "$1");
    if [[ "$result" =~ "0xc" ]]; then
        echo "vfat";
    elif [[ "$result" =~ "0x83" ]]; then
        echo "ext4";
    else
        echo "";
    fi
    
}

###############################################################################
get_image_partition_offset() {
    # $1 - full path of an image file (<filename>.img)
    # $2 - partition number (1 - /boot or 2 - /root)

    local result=$(partx -s -o START -g -n "$2:$2"  "$1");
    echo "$((result*512))";
    
}




###############################################################################
get_license_statement_from_the_header_of_this_file() {
    
    local this_file_name=$0;
    
    local current_status="selection_not_started"; # "selection_not_started", "selection_active", "selection_finished"
    local license="";
    while IFS= read -r line; do #https://www.cyberciti.biz/faq/unix-howto-read-line-by-line-from-file/
        if [[ "$line" =~ ^################################*\s*$ ]] ; then
            case $current_status in
              "selection_not_started")
                    current_status="selection_active";
                    ;;
              "selection_active")                  
                    current_status="selection_finished";
                    license+="$line";
                    ;;
              "selection_finished")      
                    ;;
              *)                  
                    ;;
            esac    
        fi
        
        if [[ "$current_status" == "selection_active" ]] ; then
            license+="$line\n";
        fi
    done < $this_file_name;

    echo -e "$license";
}


###############################################################################
license_statement() {

    echo "$(get_license_statement_from_the_header_of_this_file)";
}

###############################################################################
run_counter_onboard_support_code() {
    echo "###############################################################################
# __RUN_COUNTER__=0
###################
update_run_counter() { 
    # \$1 - increment the counter by value

    local key=\"__RUN_COUNTER__\";
    local key_rex=\"^#\s\+\$key=\";
    local value_rex=\"[0-9]\+\";
    local eol_rex=\".*$\";
    local line_rex=\"\$key_rex\$value_rex\$eol_rex\";
    local line=\$(grep -e \"\$line_rex\" \$0); 
    if [ -n \"\$line\" ] ; then
        local current_count=\$(echo \"\$line\" | cut -f2 -d= | sed 's/[^0-9]*//g'); 
        if [ \"\$1\" -gt \"0\" ] ; then
            local new_count=\$((current_count+\$1)); 
            sed -i -e \"/\$line_rex/s/\(\$key_rex\)\$value_rex\(\$eol_rex\)/\1\$new_count\2/\" \$0; 
        fi
        # return current count value
        echo \$current_count;
    else
        # return empty string
        echo \"\";
    fi
}

check_if_run_allowed__prompt_user_in_doubt__update_counter_on_yes_if_requested() { 
    # \$1 - increment the counter by value
    local current_count=\$(update_run_counter 0);
    if [ -n \"\$current_count\" ] ; then
        if [ \"\$current_count\" -gt \"0\" ] ; then
            read -p \"This script has already run \$current_count time(s). Do you really want to run it one more time? [yes[/]] \"  answer
            local yes_pat=\"^[Yy][Ee][Ss]\$\";
            if [[ \$answer =~ \$yes_pat ]] ; then
                if [ \"\$1\" -gt \"0\" ] ; then
                    update_run_counter \"$1\";
                fi
                return 0;
            fi
            
            return 1;
        fi
    fi
    return 0;
}
    
    ";
}

###############################################################################
onboard_run_counter_check() {
    # $1 - increment the counter by value

    echo "
if ! check_if_run_allowed__prompt_user_in_doubt__update_counter_on_yes_if_requested $1; then
    exit 1;
fi
";

}

###############################################################################
onboard_run_counter_check_no_increment() {

    onboard_run_counter_check "0";
}

###############################################################################
onboard_run_counter_do_increment_by_one() {

    echo "update_run_counter 1;";
}

###############################################################################
onboard_generated_automatically_message() {

    local name=$(basename "$0");
    echo "###############################################################################
# This script was generated automatically by 
# '$name' 
# at $(date -u).
###############################################################################";

}

###############################################################################
onboard_bash_script_header() {

    echo "#!/bin/bash
$(license_statement)
$(onboard_generated_automatically_message)";

}

###############################################################################
onboard_conf_bash_script_header() {

    onboard_bash_script_header;
    echo "$(run_counter_onboard_support_code)";
}

###############################################################################
###############################################################################

###############################################################################
escape_special_symbols_for_sed() {
    # $1 -- input string
    # Escapes symbols: '\', '(', ')', '?', '+'

    echo "$1" | sed -e 's/(/\\(/g' -e 's/)/\\)/g' -e 's/?/\\?/g' -e 's/+/\\+/g' -e 's/\//\\\//g' -e 's/|/\\|/g';
}

###############################################################################
get_sed_search_string() {
    # $1 -- mountpoint
    # $2 -- fstype
    # NOTE. This implementation assumes 'default' always present as the first option.
    # Groups:               (1........................1)(2......2)+(3)?(4..................4)?(5)?(6..................6)?(7)?(8..................8)?(9.9)
    local sed_search_draft="(^PARTUUID=\S+\s+$1\s+$2\s+)(defaults)+(,)?(r[wo]|noatime|noload)?(,)?(r[wo]|noatime|noload)?(,)?(r[wo]|noatime|noload)?(.*$)";
    echo $(escape_special_symbols_for_sed "$sed_search_draft");
}

###############################################################################
get_sed_replacement_string() {
    # $1 -- mountpoint
    # $2 -- fstype
    # $3 -- requested system mode: 'ro' or 'rw'.

    if [[ $3 = "ro" ]] ; then
        if [[ $1 =~ "/boot" ]]; then
            echo ",ro,noatime";
        else
            echo ",ro,noatime,noload";
        fi
    elif  [[ $3 = "rw" ]] ; then
        # "rw" mode
        if [[ $1 =~ "/boot" ]]; then
            echo "";    #",rw";
        else
            echo ",noatime";    #",rw,noatime";
        fi
    else 
        echo ",ERROR_UNKNOWN_SYSTEM_MODE_REQUESTED_$3";
    fi
}

###############################################################################
get_sed_replace_command() {
    # $1 -- mountpoint
    # $2 -- fstype
    # $3 -- requested system mode: 'ro' or 'rw'.
    
    local sed_arg_search=$(get_sed_search_string $1 $2);
    local sed_arg_replacement=$(get_sed_replacement_string $1 $2 $3);
    
    echo "s/$sed_arg_search/\1\2$sed_arg_replacement\9/";
}


###############################################################################
###############################################################################


###############################################################################
onboard_backup_etc_fstab() {

    local fstab_file_path="/etc/fstab";
    
    echo "
if [[ -e $fstab_file_path.bak0 ]] ; then
    cp $fstab_file_path $fstab_file_path.bak;
else
    cp $fstab_file_path $fstab_file_path.bak0;
fi
";
    
}

###############################################################################
onboard_modify_etc_fstab_to_boot_in_rw_or_ro_mode() {
    # $1 - option code
    #       "1" or "rw"         - "rw"
    #       any other or empty  - "ro" (default) 

    local boot_mode="ro";
    local fstab_file_path="/etc/fstab";
    
    case $1 in
      "1")      boot_mode="rw";         ;;
      "rw")     boot_mode="rw";         ;;
    esac    
    
    echo "sed -i -e '$(get_sed_replace_command /boot vfat $boot_mode)' -e '$(get_sed_replace_command / ext4 $boot_mode)'  $fstab_file_path";
    
}


###############################################################################
onboard_the_end_message() {

    echo "
###############################################################################
###################  THE END  #################################################
###############################################################################
";

}


###############################################################################
terminate() { echo "$1. The script is terminated."; exit 1; }

###############################################################################
return_failed() { echo "$1."; return 1; }

###############################################################################
is_root() { return $(ls /root/ &>/dev/null); }

###############################################################################
exit_if_not_root() {
    if ! is_root; then
        terminate "Please run this script as root (use sudo)";
    fi
}

###############################################################################
onboard_is_root_check_code() {

    echo "
is_root() { return \$(ls /root/ &>/dev/null); }

if ! is_root; then
    echo \"Please run this script as root (use sudo)\";
    exit 1;
fi
";

}

###############################################################################
onboard_report_support() {

    echo "

onboard_create_report() {
    # \$1 - command
    # \$2 - command nickname
    # \$3 - output directory (without closing slash)
    # \$4 - output file name prefix
    # \$5 - output file name postfix
    # \$6 - owner name
    # \$7 - Optional. A string to output instead of the command ($1) report.
    # Returns:
    #       echo \"\$errors\"

    local errors=\"\";
    local path=\"\$3/\$4_\$2_\$5_report.txt\";

    local true_str=\"TRUE\";
    local ok=\"\";


    if [[ -n \$7 ]] ; then
        if echo \"\$7\" > \$path; then
            ok=\$true_str;
        fi
    else
        if \$1 > \$path; then
            ok=\$true_str;
        fi
    fi

    if [[ -n \$ok ]] ; then
        if( ! chown \$6:\$6 \$path ) ; then
            errors+=\"\$2-chown;\";
        fi
        if( ! chmod 644 \$path ) ; then
            errors+=\"\$2-chmod;\";
        fi
    else
        errors+=\"\$2;\";
    fi    

    echo \$errors;
}

";

}


###############################################################################
onboard_mount_support() {

    echo "
is_system_mounted_as_rw() {
    # returns  TRUE if none of '/' and '/boot/' found out to be 'ro'.
    # returns  FALSE otherwise.
    return \$( ! findmnt -lo target,options | grep -e '^/\s\+ro' -e '^/boot.*\s\+ro' &>/dev/null);
}

is_system_root_mounted_as_rw() {
    # returns  TRUE if '/' is 'rw'.
    # returns  FALSE otherwise.
    return \$( findmnt -lo target,options | grep -e '^/\s\+rw' &>/dev/null);
}

is_system_boot_mounted_as_rw() {
    # returns  TRUE if '/boot/' is 'rw'.
    # returns  FALSE otherwise.
    return \$( findmnt -lo target,options | grep -e '^/boot.*\s\+rw' &>/dev/null);
}

";

}




###############################################################################
is_in_graphical_target() { return $(systemctl get-default | grep "graphical.target" &>/dev/null); }

#######################################################################################
error_exit()
    # $1 - line number -- ${LINENO}
    # $2 - message
{
    local err=$?;
    if [ "$err" != "0" ]; then
        echo "Line: $1  $2"
        exit 1
    fi
}


###############################################################################
check_if_zip_file_exists_and_exit_if_not() {
    # $1 - zip_file_full_path
    
    if [ -e "$1" ]; then
        if [ -r "$1" ]; then
            # Check size of the file
            local file_size=$(wc -c "$1" | cut -d' ' -f 1);
                echo "Zip file '.../$(basename "$1")' with original image exists, has size of $file_size bytes, and is readable"
        else
            terminate "Zip file with original image '$1' is not readable by you";
        fi
    else
        terminate "Provided path to Zip file with original image does not exist: '$1'";
    fi

}

###############################################################################
check_if_img_file_exists_and_exit_if_not() {
    # $1 - img_file_full_path
    
    if [ -e "$1" ]; then
        if [ -r "$1" ]; then
            return 0; 
        else
            terminate "Image file '$1' is not readable by you";
        fi
    else
        terminate "Provided path to image file does not exist: '$1'";
    fi

}

###############################################################################
unzip_image_to_tmp_image_dir() {
    # $1 - zip_file_name
    # $2 - tmp_image_dir
    echo "Unzipping original image"
    if 7z e "$1" -o"$2"; then
        return 0;
    else
        terminate "Can't unzip original image";
    fi
}

###############################################################################
mount_original_image_to_tmp_boot_dir() {
    # $1 - img_file_full_path
    # $2 - tmp_mount_boot_dir
    
    local image_path="$1";
    local part_number=1;
    local part_type=$(get_image_partition_type "$image_path" part_number);
    local part_offset=$(get_image_partition_offset "$image_path" part_number);
    local tmp_mount_dir=$2;
    echo "Image Path:           $image_path";
    echo "Boot Partition Type:      $part_type";
    echo "Boot Partition Offset:    $part_offset";
    echo "Boot Mount Point:     $tmp_mount_dir";

    if mount -v -t "$part_type" -o loop,offset="$part_offset" "$image_path" "$tmp_mount_dir"; then
        echo "Image Boot partition mounted successfully"
        return 0;
    else
        return_failed "Can't mount original image Boot partition";
        return 1;
    fi

}

###############################################################################
mount_original_image_to_tmp_root_dir() {
    # $1 - img_file_full_path
    # $2 - tmp_mount_root_dir
    
    local image_path="$1";
    local part_number=2;
    local part_type=$(get_image_partition_type "$image_path" "$part_number");
    local part_offset=$(get_image_partition_offset "$image_path" "$part_number");
    local tmp_mount_dir=$2;
    echo "Image Path:           $image_path";
    echo "Root Partition Type:      $part_type";
    echo "Root Partition Offset:    $part_offset";
    echo "Root Mount Point:     $tmp_mount_dir";

    if mount -v -t "$part_type" -o loop,offset="$part_offset" "$image_path" "$tmp_mount_dir"; then
        echo "Image Root partition mounted successfully"
        return 0;
    else
        return_failed "Can't mount original image Root partition";
        return 1;
    fi

}


###############################################################################
mount_original_image_and_prepare_root_for_read_only_mode() {
    # $1 - img_file_full_path
    # $2 - tmp_mount_root_dir

    local img_file_full_path=$1;
    local tmp_mount_root_dir=$2;
    if mount_original_image_to_tmp_root_dir "$img_file_full_path" "$tmp_mount_root_dir"; then {
    
        echo "Prepare script to create initial boot reports."
        create_script_to_collect_initial_boot_reports "$tmp_mount_root_dir" "conf-0-1-"

        echo "Prepare script for ufw install and default setup."
        create_script_for_ufw_install_enable_and_reboot "$tmp_mount_root_dir" "conf-0-2-"

        echo "Prepare script for regular updates with automated switching to RW and reboot if succeeded."
        create_script_to_make_full_system_update "$tmp_mount_root_dir" "conf-0-3-"

        echo "Prepare script to create reports after system upgrade."
        create_script_to_collect_reports_after_system_upgrade "$tmp_mount_root_dir" "conf-0-4-"

        echo "Prepare script to configure system read-only mode."
        create_script_to_configure_system_read_only_mode "$tmp_mount_root_dir" "conf-0-5-"
        
        echo "Prepare script to permanently switch to RO mode (via fstab) and reboot if succeeded."
        create_script_to_modify_etc_fstab_to_switch_permanently_to_read_only_mode_and_reboot "$tmp_mount_root_dir" "conf-0-6-"
        
        echo "Prepare script to create reports after system switching to read only mode."
        create_script_to_collect_reports_after_switching_to_read_only_mode "$tmp_mount_root_dir" "conf-0-7-"

        echo "Prepare scripts to switch between RO & RW modes."
        create_scripts_to_switch_temporarily_between_RO_and_RW_modes "$tmp_mount_root_dir"
    
        echo "Prepare scripts to switch permanently (default boot) between RW and RO modes via /etc/fstab."
        create_script_to_modify_etc_fstab_to_boot_in_mode "$tmp_mount_root_dir" "0"
        create_script_to_modify_etc_fstab_to_boot_in_mode "$tmp_mount_root_dir" "1"
        
        echo "Prepare script for regular updates with automated switching to RW and reboot if succeeded."
        create_script_to_make_system_regular_update_and_reboot "$tmp_mount_root_dir" ""
        
        echo "Prepare script to create regular reports."
        create_script_to_collect_reports "$tmp_mount_root_dir" ""

        return 0;
    }
    else
        return 1;
    fi


}


###############################################################################
create_script_to_collect_initial_boot_reports() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-1-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2collect-initial-boot-reports.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

$(onboard_report_support)

do_reports() {

    local error=\"\";

    error+=\$(onboard_create_report \"df -h\" \"df\" \"/home/pi\" \"01\" \"initial_boot\" \"pi\");
    error+=\$(onboard_create_report \"free\" \"free\" \"/home/pi\" \"02\" \"initial_boot\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b -p 0..3\" \"journal_b_p0_3\" \"/home/pi\" \"03\" \"initial_boot\" \"pi\");
    error+=\$(onboard_create_report \"aplay -l\" \"aplay\" \"/home/pi\" \"04\" \"initial_boot\" \"pi\");
    error+=\$(onboard_create_report \"ls -lsh /usr/local/sbin\" \"ls_usr_local_sbin\" \"/home/pi\" \"05\" \"initial_boot\" \"pi\");


    if [[ -z \$error ]] ; then
        ls -lsh /home/pi/0*;
        echo -e \"\n\t\t\t\tInitial boot reports created SUCCESSFULLY.\";
        $(onboard_run_counter_do_increment_by_one)
    else
        echo \"FAILED to create initial boot reports: '\$error'. See the errors above.\";
    fi
}

do_reports;

$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}




###############################################################################
create_script_for_ufw_install_enable_and_reboot() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-2-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2ufw-install-enable-and-reboot.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

if apt update --assume-yes ; then
    if apt install --assume-yes ufw; then
        if ufw enable ; then
            if ufw status verbose ; then 
                echo -e \"\n\n\n\t\t\t\tUFW installed and enabled SUCCESSFULLY. Reboot in 10 seconds...\";
                $(onboard_run_counter_do_increment_by_one)
                sleep 10;
                reboot;
                exit 0;
            fi
        fi
    fi    
fi

echo \"Upgrade FAILED. See the errors above.\";
    
$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}

###############################################################################
remount_mountpoint() {
    # $1 -- mountpoint (For example, "/" - for root, "/boot" - for boot)
    # $2 -- mode to remount to ( "ro", "rw")

    local grep_filter="grep -w ^$1";
    local current_mount_info=$(findmnt -u -o TARGET,SOURCE,FSTYPE -D | $grep_filter);

    #https://www.gnu.org/software/sed/manual/sed.html
    local sed_search_param_1=$(escape_special_symbols_for_sed "^($1\S*)\s*.*$");
    local mountpoint=$(echo "$current_mount_info" | sed -e "s/$sed_search_param_1/\1/");

    local sed_search_param_2=$(escape_special_symbols_for_sed "^($1\S*)\s*(\S*)\s*.*$");
    local device=$(echo "$current_mount_info" | sed -e "s/$sed_search_param_2/\2/");
    
    local sed_search_param_3=$(escape_special_symbols_for_sed "^($1\S*)\s*(\S*)\s*(\S*)\s*.*$");
    local fstype=$(echo "$current_mount_info" | sed -e "s/$sed_search_param_3/\3/");

    local sed_replacement_string=$(get_sed_replacement_string "$mountpoint" "$fstype" "$2");

    local remount_options="remount,defaults";
    remount_options+=$(get_sed_replacement_string "$mountpoint" "$fstype" "$2");

    mount -o $remount_options $device $mountpoint ;
    return $?;

}

###############################################################################
remount_boot_to_ro() {
    remount_mountpoint "/boot" "ro";
    return $?;
}

###############################################################################
remount_boot_to_rw() {
    remount_mountpoint "/boot" "rw";
    return $?;
}

###############################################################################
remount_root_to_ro() {
    remount_mountpoint "/" "ro";
    return $?;
}

###############################################################################
remount_root_to_rw() {
    remount_mountpoint "/" "rw";
    return $?;
}

###############################################################################
onboard_remount_mountpoint_support() {

    echo "
escape_special_symbols_for_sed() {
    # \$1 -- input string
    # Escapes symbols: '\', '(', ')', '?', '+'

    echo \"\$1\" | sed -e 's/(/\\\\(/g' -e 's/)/\\\\)/g' -e 's/?/\\\\?/g' -e 's/+/\\\\+/g' -e 's/\//\\\\\//g' -e 's/|/\\\\|/g';
}

get_sed_replacement_string() {
    # \$1 -- mountpoint
    # \$2 -- fstype
    # \$3 -- requested system mode: 'ro' or 'rw'.

    if [[ \$3 = \"ro\" ]] ; then
        if [[ \$1 =~ \"/boot\" ]]; then
            echo \",ro,noatime\";
        else
            echo \",ro,noatime,noload\";
        fi
    elif  [[ \$3 = \"rw\" ]] ; then
        # \"rw\" mode
        if [[ \$1 =~ \"/boot\" ]]; then
            echo \"\";    #\",rw\";
        else
            echo \",noatime\";    #\",rw,noatime\";
        fi
    else 
        echo \",ERROR_UNKNOWN_SYSTEM_MODE_REQUESTED_\$3\";
    fi
}


    
remount_mountpoint() {
    # \$1 -- mountpoint (For example, \"/\" - for root, \"/boot\" - for boot)
    # \$2 -- mode to remount to ( \"ro\", \"rw\")

    local grep_filter=\"grep -w ^\$1\";
    local current_mount_info=\$(findmnt -u -o TARGET,SOURCE,FSTYPE -D | \$grep_filter);

    #https://www.gnu.org/software/sed/manual/sed.html
    local sed_search_param_1=\$(escape_special_symbols_for_sed \"^(\$1\S*)\s*.*\$\");
    local mountpoint=\$(echo \"\$current_mount_info\" | sed -e \"s/\$sed_search_param_1/\1/\");

    local sed_search_param_2=\$(escape_special_symbols_for_sed \"^(\$1\S*)\s*(\S*)\s*.*\$\");
    local device=\$(echo \"\$current_mount_info\" | sed -e \"s/\$sed_search_param_2/\2/\");
    
    local sed_search_param_3=\$(escape_special_symbols_for_sed \"^(\$1\S*)\s*(\S*)\s*(\S*)\s*.*\$\");
    local fstype=\$(echo \"\$current_mount_info\" | sed -e \"s/\$sed_search_param_3/\3/\");

    local sed_replacement_string=\$(get_sed_replacement_string \"\$mountpoint\" \"\$fstype\" \"\$2\");

    local remount_options=\"remount,defaults\";
    remount_options+=\$(get_sed_replacement_string \"\$mountpoint\" \"\$fstype\" \"\$2\");

    mount -o \$remount_options \$device \$mountpoint ;
    #echo \"ECHO: 'mount -o \$remount_options \$device \$mountpoint ;'\"
    return \$?;

}

remount_boot_to_ro() {
    remount_mountpoint \"/boot\" \"ro\";
    return \$?;
}

remount_boot_to_rw() {
    remount_mountpoint \"/boot\" \"rw\";
    return \$?;
}

remount_root_to_ro() {
    remount_mountpoint \"/\" \"ro\";
    return \$?;
}

remount_root_to_rw() {
    remount_mountpoint \"/\" \"rw\";
    return \$?;
}


";

}

###############################################################################
create_script_to_make_full_system_update() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-3-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2update-system-software-and-reboot.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

$(onboard_remount_mountpoint_support)

if remount_root_to_rw ; then
    echo \"Mounted '/' as RW.\";
    if remount_boot_to_rw ; then
        echo \"Mounted '/boot/' as RW.\";
        echo \"Updating & Upgrading... Please wait ...\";
        if apt update --assume-yes ; then
            if apt upgrade --assume-yes ; then
                echo -e \"\n\n\n\t\t\t\tUpgrade SUCCEEDED. Reboot in 10 seconds...\";
                $(onboard_run_counter_do_increment_by_one)
                sleep 10;
                reboot;
                exit 0;
            fi
        fi    
    fi    
fi

echo \"Upgrade FAILED. See the errors above.\";
    
$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}


###############################################################################
create_script_to_collect_reports_after_system_upgrade() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-4-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2collect-reports-after-system-upgrade.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

$(onboard_report_support)

do_reports() {

    local error=\"\";

    error+=\$(onboard_create_report \"df -h\" \"df\" \"/home/pi\" \"11\" \"after_system_upgrade\" \"pi\");
    error+=\$(onboard_create_report \"free\" \"free\" \"/home/pi\" \"12\" \"after_system_upgrade\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b -p 0..3\" \"journal_b_p0_3\" \"/home/pi\" \"13\" \"after_system_upgrade\" \"pi\");
    error+=\$(onboard_create_report \"aplay -l\" \"aplay\" \"/home/pi\" \"14\" \"after_system_upgrade\" \"pi\");


    if [[ -z \$error ]] ; then
        ls -lsh /home/pi/1*;
        echo -e \"\n\t\t\t\tReports after system upgrade created SUCCESSFULLY.\";
        $(onboard_run_counter_do_increment_by_one)
    else
        echo \"FAILED to create reports after system upgrade: '\$error'. See the errors above.\";
    fi
}

do_reports;
    
$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}




###############################################################################
create_script_to_configure_system_read_only_mode() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-5-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2prepare-system-read-only-mode.sh";
    error_exit ${LINENO} "Failed to initialize 'script_ro_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

disable_swap() {
    systemctl stop dphys-swapfile
    systemctl mask dphys-swapfile
    rm /var/swap
}



create_tmpfs_unit_based_on_tmp_mount() {
    # arg 1 - where (mount point dir without closing slash)
    # arg 2 - options (for example, mode=1777,strictatime,nosuid,nodev; or mode=755,relatime,nosuid,nodev)
    #       if empty it is not changed.
    local template_unit_where=\"/tmp\";
    local template_path=\"/usr/share/systemd/tmp.mount\";
    local template_ext=\"\${template_path##*.}\"
    local new_unit_dir=\"/etc/systemd/system\";
    local new_unit_where=\$1;
    local new_unit_name=\$(echo \"\$new_unit_where\" | sed \"s/\/*//\" | sed \"s/\//-/g\");
    local new_unit_path=\"\$new_unit_dir/\$new_unit_name.\$template_ext\";
    cp \$template_path \$new_unit_path

    if [[ ! -z \$2 ]]; then 
        sed -i -e \"/Description=/s@/tmp@\$new_unit_where@\" -e \"/ConditionPathIsSymbolicLink=/s@/tmp@\$new_unit_where@\" -e \"/Where=/s@/tmp@\$new_unit_where@\" -e \"/Options=/s@Options=.*@Options=\$2@\"  \$new_unit_path; 
    else 
        sed -i -e \"/Description=/s@/tmp@\$new_unit_where@\" -e \"/ConditionPathIsSymbolicLink=/s@/tmp@\$new_unit_where@\" -e \"/Where=/s@/tmp@\$new_unit_where@\" \$new_unit_path; 
    fi

    sed -i \"1i\\
###############################################################################\\n\\
# This script is a modified copy automatically created\\n\\
# from the system template '\$template_path'\\n\\
# by '\$\(basename \"\$0\"\)'\\n\\
# at \$\(date -u\).\\n\\
###############################################################################\\n\" \$new_unit_path


}

create_tmpfs_units() {

    create_tmpfs_unit_based_on_tmp_mount \"/tmp\"      # keep default /tmp mode 'as-is'
    create_tmpfs_unit_based_on_tmp_mount \"/var/tmp\"  # keep default /tmp mode 'as-is'
    create_tmpfs_unit_based_on_tmp_mount \"/var/log\" \"mode=755,relatime,nosuid,nodev\"         # set regular mode
    create_tmpfs_unit_based_on_tmp_mount \"/var/lib/dhcp\" \"mode=755,relatime,nosuid,nodev\"    # set regular mode
    create_tmpfs_unit_based_on_tmp_mount \"/var/lib/dhcpcd5\" \"mode=755,relatime,nosuid,nodev\" # set regular mode
    create_tmpfs_unit_based_on_tmp_mount \"/var/lib/systemd\" \"mode=755,relatime,nosuid,nodev\" # set regular mode
    create_tmpfs_unit_based_on_tmp_mount \"/media\" \"mode=755,relatime,nosuid,nodev\"           # set regular mode
    create_tmpfs_unit_based_on_tmp_mount \"/mnt\" \"mode=755,relatime,nosuid,nodev\"             # set regular mode
}

enable_tmpfs_units() {

    systemctl enable tmp.mount
    systemctl enable var-tmp.mount
    systemctl enable var-log.mount
    systemctl enable var-lib-dhcp.mount
    systemctl enable var-lib-dhcpcd5.mount
    systemctl enable var-lib-systemd.mount
    systemctl enable media.mount
    systemctl enable mnt.mount
}

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

disable_swap;
create_tmpfs_units;
enable_tmpfs_units;

is_in_graphical_target() { return \$(systemctl get-default | grep \"graphical.target\" &>/dev/null); }

do_switch_to_cli() {

    if is_in_graphical_target; then

        echo \"Current read-only mode implementation is not compatible with graphical target.\";        
        read -p \"Do you want to switch the system to 'Boot to CLI' mode right now? [y/] \"  answer
        local yes_pat=\"^[Yy]\$\";
        if [[ \$answer =~ \$yes_pat ]] ; then

            # Switch the system to \"Boot to CLI\" mode.
            if raspi-config nonint do_boot_behaviour B2 ; then
                echo -e \"\n\n\n\t\tThe system has been switched to 'Boot to CLI' mode. Reboot in 10 seconds...\";
                $(onboard_run_counter_do_increment_by_one)
                sleep 10;
                reboot;
            else
                echo \"Cannot switch system to 'Boot to CLI' mode. Try to do that manually.\";
            fi
        fi
    fi
}

do_switch_to_cli;
    
$(onboard_the_end_message)
    
" > $script_file_full_path;

  
    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}


###############################################################################
create_script_to_modify_etc_fstab_to_switch_permanently_to_read_only_mode_and_reboot() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-6-")

    local boot_mode="ro";
    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2switch-permanently-to-read-only-mode-and-reboot.sh";
    error_exit ${LINENO} "Failed to initialize 'script_ro_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

modify_fstab_and_reboot() {
    $(onboard_backup_etc_fstab)
    if $(onboard_modify_etc_fstab_to_boot_in_rw_or_ro_mode $boot_mode) ; then
        echo -e \"\n\n\n\t\tThe system has been switched to '$boot_mode' boot mode. Reboot in 10 seconds...\";
        $(onboard_run_counter_do_increment_by_one)
        sleep 10;
        reboot;
    else
        echo \"Cannot modify fstab file to switch to '$boot_mode' boot mode.\"    
    fi
}

is_in_graphical_target() { return \$(systemctl get-default | grep \"graphical.target\" &>/dev/null); }

report_incompatibility_with_graphical_mode_and_exit() {

    echo \"Current read-only mode implementation is not compatible with graphical target. Please switch the system to 'Boot to CLI' mode, then reboot and run this script again.\";
        exit 1;
}

if is_in_graphical_target; then

    echo \"Current read-only mode implementation is not compatible with graphical target.\";        
    read -p \"Do you want to switch the system to 'Boot to CLI' mode right now? [y/] \"  answer
    local yes_pat=\"^[Yy]\$\";
    if [[ \$answer =~ \$yes_pat ]] ; then

        if raspi-config nonint do_boot_behavior B2 ; then
            modify_fstab_and_reboot;
        else
            echo \"Cannot switch system to 'Boot to CLI' mode. Try to do that manually.\";
        fi

    else
        report_incompatibility_with_graphical_mode_and_exit;
    fi
else
    modify_fstab_and_reboot;
fi

$(onboard_the_end_message)
    
" > $script_file_full_path;
   
    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";

}

###############################################################################
create_script_to_collect_reports_after_switching_to_read_only_mode() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("conf-0-7-")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2collect-reports-after-switching-to-read-only-mode-and-reboot.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_conf_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_run_counter_check_no_increment)

$(onboard_mount_support)

$(onboard_report_support)

$(onboard_remount_mountpoint_support)

do_reports_and_reboot() {

    local error=\"\";

    local old_is_ro=\"\";
    if ! is_system_root_mounted_as_rw ; then
        old_is_ro=\"TRUE\";
        if ! remount_root_to_rw ; then
            local error+=\"mount_rw;\";
        fi
    fi

    error+=\$(onboard_create_report \"df -h\" \"df\" \"/home/pi\" \"21\" \"after_switching_to_read_only_mode\" \"pi\");
    error+=\$(onboard_create_report \"free\" \"free\" \"/home/pi\" \"22\" \"after_switching_to_read_only_mode\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b -p 0..3\" \"journal_b_p0_3\" \"/home/pi\" \"23\" \"after_switching_to_read_only_mode\" \"pi\");
    error+=\$(onboard_create_report \"aplay -l\" \"aplay\" \"/home/pi\" \"24\" \"after_switching_to_read_only_mode\" \"pi\");

    if ! ls -lsh /home/pi/2*; then
        error+=\"ls;\";
    fi

    if [[ -z \$error ]] ; then
        echo -e \"\n\t\t\t\tReports for system running in read-only mode created SUCCESSFULLY. Reboot in 10 seconds...\";
        $(onboard_run_counter_do_increment_by_one)
        sleep 10;
        reboot;
    else
        echo \"FAILED during report creation. The reports may be created OK. Hint for errors: '\$error'.\";
    fi
}

do_reports_and_reboot;

$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}

###############################################################################
create_script_to_collect_reports() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    #local script_file_name="$2collect-reports-and-reboot.sh";
    #local script_file_name="$2regular-reports-collect.sh";
    local script_file_name="$2reports-collect.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";



    echo "$(onboard_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_mount_support)

$(onboard_report_support)

$(onboard_remount_mountpoint_support)

do_reports() {

    local error=\"\";

    local findmnt_report=\"\";    
    
    local old_is_ro=\"\";
    if ! is_system_root_mounted_as_rw ; then
        old_is_ro=\"TRUE\";
        findmnt_report=\$(findmnt);    
        if ! remount_root_to_rw ; then
            error+=\"mount_rw;\";
        fi
    fi

    local id=\"5\";
    local timestamp=\$(date +%Y-%m-%d_%H%M%S);
    local prefix=\$timestamp\"__\$id\";
    local dir=\"/home/pi\";
    error+=\$(onboard_create_report \"df -h\" \"df\" \"\$dir\" \$prefix'01_' \"_\" \"pi\");
    error+=\$(onboard_create_report \"free\" \"free\" \"\$dir\" \$prefix'02_' \"_\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b -p 0..3\" \"journal_b_p0_3\" \"\$dir\" \$prefix'03_' \"_\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b -p 0..4\" \"journal_b_p0_4\" \"\$dir\" \$prefix'04_' \"_\" \"pi\");
    error+=\$(onboard_create_report \"journalctl -b\" \"journal_b\" \"\$dir\" \$prefix'05_' \"_\" \"pi\");
    if [[ -n \$old_is_ro ]] ; then
        error+=\$(onboard_create_report \"\" \"findmnt\" \"\$dir\" \$prefix'06_' \"_\" \"pi\" \"\$findmnt_report\");
    else
        error+=\$(onboard_create_report \"findmnt\" \"findmnt\" \"\$dir\" \$prefix'06_' \"_\" \"pi\");
    fi
    error+=\$(onboard_create_report \"lsblk -o NAME,RO,TYPE,FSTYPE,MOUNTPOINT,LABEL\" \"lsblk\" \"\$dir\" \$prefix'07_' \"_\" \"pi\");
    #error+=\$(onboard_create_report \"systemctl status mpd\" \"mpd_status\" \"\$dir\" \$prefix'08_' \"_\" \"pi\");
    #error+=\$(onboard_create_report \"mpc listall\" \"mpc_listall\" \"\$dir\" \$prefix'09_' \"_\" \"pi\");

    if ! ls -lsh \$dir/\$prefix*; then
        error+=\"ls;\";
    fi

    if [[ -n \$old_is_ro ]] ; then
        if ! remount_root_to_ro ; then
            error+=\"mount_ro_back;\";
        fi
    fi


    if [[ -z \$error ]] ; then

        echo -e \"\n\t\t\t\tReports created SUCCESSFULLY.\";

    else
        echo \"FAILED during report creation. The reports may be created OK. Hint for errors: '\$error'.\";
    fi
}

do_reports;

$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}



###############################################################################
create_scripts_to_switch_temporarily_between_RO_and_RW_modes() {
    # $1 - file system root mount point (dir without closing slash)

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_ro_file_name="ro-remount.sh";
    error_exit ${LINENO} "Failed to initialize 'script_ro_file_name' local variable.";
    local script_rw_file_name="rw-remount.sh";
    error_exit ${LINENO} "Failed to initialize 'script_rw_file_name' local variable.";
    local script_ro_file_full_path="$1$scripts_root_dir/$script_ro_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_ro_file_full_path' local variable.";
    local script_rw_file_full_path="$1$scripts_root_dir/$script_rw_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_rw_file_full_path' local variable.";

    echo "$(onboard_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_remount_mountpoint_support)

remount_root_to_ro;
remount_boot_to_ro;
    
$(onboard_the_end_message)
    
" > $script_ro_file_full_path;
   
    error_exit ${LINENO} "Failed to create file '$script_ro_file_full_path'.";

    chown 1000:1000 $script_ro_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_ro_file_full_path'.";
    chmod 744 $script_ro_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_ro_file_full_path'.";


    echo "$(onboard_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_remount_mountpoint_support)

remount_root_to_rw;
remount_boot_to_rw;
    
$(onboard_the_end_message)
    
" > $script_rw_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_rw_file_full_path'.";

    chown 1000:1000 $script_rw_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_rw_file_full_path'.";
    chmod 744 $script_rw_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_rw_file_full_path'.";
}


###############################################################################
create_script_to_modify_etc_fstab_to_boot_in_mode() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - option code
    #       1                   - "rw"
    #       any other or empty  - "ro" (default) 

    local boot_mode="ro";
    if [[ "$2" == "1" ]]; then
        boot_mode="rw";
    fi 
    local script_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'script_root_dir' local variable.";
    local script_file_name="switch-boot-to-$boot_mode.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$script_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_bash_script_header)

$(onboard_mount_support)

$(onboard_is_root_check_code)

$(onboard_remount_mountpoint_support)

old_is_ro=\"\";
if ! is_system_root_mounted_as_rw ; then
    old_is_ro=\"TRUE\";
    if ! remount_root_to_rw ; then
        echo \"Cannot remount root to 'rw' mode.\";
        exit 1;
    fi
fi

$(onboard_backup_etc_fstab)
$(onboard_modify_etc_fstab_to_boot_in_rw_or_ro_mode $boot_mode);

echo \"The system has been switched to the '$boot_mode' boot mode.\";
echo \"Note that the changes will only take effect after reboot.\";

if [[ -n \$old_is_ro ]] ; then
    if ! remount_root_to_ro ; then
        echo \"Cannot remount root back to 'ro' mode.\";
    fi
fi

$(onboard_the_end_message)
    
" > $script_file_full_path;
   
    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";

}

###############################################################################
create_script_to_make_system_regular_update_and_reboot() {
    # $1 - file system root mount point (dir without closing slash)
    # $2 - script file name prefix ("")

    local scripts_root_dir="/usr/local/sbin";
    error_exit ${LINENO} "Failed to initialize 'scripts_root_dir' local variable.";
    local script_file_name="$2system-regular-update-and-reboot.sh";
    error_exit ${LINENO} "Failed to initialize 'script_file_name' local variable.";
    local script_file_full_path="$1$scripts_root_dir/$script_file_name";
    error_exit ${LINENO} "Failed to initialize 'script_file_full_path' local variable.";

    echo "$(onboard_bash_script_header)

$(onboard_is_root_check_code)

$(onboard_remount_mountpoint_support)

if remount_root_to_rw ; then
    echo \"Mounted '/' as RW.\";
    if remount_boot_to_rw ; then
        echo \"Mounted '/boot/' as RW.\";
        echo \"Updating & Upgrading... Please wait ...\";
        if apt update --assume-yes ; then
            if apt upgrade --assume-yes ; then
                echo -e \"\n\n\n\t\t\t\tUpgrade SUCCEEDED. Reboot in 10 seconds...\";
                sleep 10;
                reboot;
                exit 0;
            fi
        fi    
    fi    
fi

echo \"Upgrade FAILED. See the errors above.\";
    
$(onboard_the_end_message)
    
" > $script_file_full_path;

    error_exit ${LINENO} "Failed to create file '$script_file_full_path'.";

    chown 1000:1000 $script_file_full_path;
    error_exit ${LINENO} "Failed to set owner & group for file '$script_file_full_path'.";
    chmod 744 $script_file_full_path;
    error_exit ${LINENO} "Failed to set executable permissions for file '$script_file_full_path'.";
}



#######################################################################################
DEBUG_wait_press_any_key() {

    echo "Press any key to continue... [$(date +%H:%M:%S)]"
    while [ true ] ; do
        read -t 10 -n 1
        if [ $? = 0 ] ; then
            echo " ==> key pressed @ [$(date +%H:%M:%S)] ==> Stopped waiting."
            break;  #return ;
        else
            echo "... waiting for the keypress [$(date +%H:%M:%S)]"
        fi
    done
}

###############################################################################
move_modified_system_image() {
    # $1 - full path of an image file (/dirpath/<filename>.img) to move
    # $2 - destination directory to store modified RPi image (dir without closing slash) for further operations

    mv "$1" "$2" 

}

###############################################################################
include_custom_additions() {

    ####################################################################
    # NOTE.
    # More "custom-additions-*" to come. 
    # Please visit https://altomaxtech.com/ often to check out the news.
    ####################################################################

    local files_mask="*custom-additions-*";
    local custom_additions_list="";
    if ls -1 $files_mask  &>/dev/null ; then
        custom_additions_list=$(ls -1 $files_mask);
        for custom_additions_item in $custom_additions_list
        do
            if [[ -r $custom_additions_item ]] ; then
                source $custom_additions_item
            fi
        done
    fi
}


#################################################################################
#################################################################################
#################################################################################


