#!/bin/bash
###############################################################################
# This library "prepare-system-image-for-read-only-mode" helps to configure 
# an image of Raspberry Pi OS or other compatible Linux distributions 
# to run in read-only mode.
#
# This library is supplementary material to the article 
# "How to configure Raspberry Pi OS to run in read-only mode". 
# The article and this library can be found at https://altomaxtech.com/. 
#
# Copyright (C) 2021, Vladimir Zlatov, Evgueni Tzvetanov, USA 
# Email us at support@altomaxtech.com. 
# 
# This library is free software; you can redistribute it 
# and/or modify it under the terms of the GNU Lesser General Public 
# License as published by the Free Software Foundation; either 
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General 
# Public License along with this library; 
# if not, write to the Free Software Foundation, Inc., 
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
###############################################################################
###############################################################################
# Notes. 
#    This library is able to automatically find Raspberry Pi OS zip files 
#    with images using observed current naming scheme (for example, like 
#    "2021-05-07-raspios-buster-armhf.zip" for Raspberry Pi OS with Desktop).
#    There is no guarantee that this scheme will be used in future releases 
#    of Raspberry Pi OS Team. 
#    The naming scheme is defined by get_default_rpi_os_filename_regex function.
#    The Raspberry Pi OS images can be downloaded from the developer's website:
#    https://www.raspberrypi.org/software/operating-systems/
#    This Library is suitable for any of the packages available there:
#       - Raspberry Pi OS with desktop and recommended software; 
#       - Raspberry Pi OS with desktop;
#       - Raspberry Pi OS Lite.
######################################################################
#    This library uses the /tmp directory for all intermediate files and mounts.
#
###############################################################################
# Known issues:
#   - In case of unexpected errors the script may terminate without proper  
#     unmounting of currently mounted partition(s) of the image.
#     The incompleted image and temporary mount directories can be dismissed 
#     assuming the system automatic cleaning care of temporary locations.
#     Or, the image and the directories can be removed manually if needed.
#   - The error handling is not consistent enough in some operators which may 
#     allow next command(s) to proceed as normal without noticing that 
#     the previous command has failed.
#     It is always a good idea to check out reports for errors.
#      
# 
###############################################################################

source "prepare-system-image-for-read-only-mode-functions-00.sh"



###############################################################################
###############################################################################
main() {
    ###########################################################################
    # $1 - original image zip file directory (without closing slash). 
    #      If empty (or ".") the current directory is assumed. 
    ###########################################################################
    # Usage:
    #     prepare-system-image-for-read-only-mode ["<path_to_original_zip_with_image>"]
    #
    ###########################################################################
    # Note.
    #     This script is capable to handle multiple image files and 
    #     will try to process all zip files in the source directory
    #     without preliminary checks of the content.
    #     It is a user responsibility to make sure that the directory
    #     contains valid image files or (sym)links to them.
    ###########################################################################

    local this_module_name="prepare-system-image-for-read-only-mode";
    local modified_image_destination_directory="/tmp";

    echo "'$this_module_name' STARTED  at $(date)" 

    exit_if_not_root;

    local arg1=$(echo "$1" | sed "s/\/\s*$//");

    local zip_file_dir=$arg1;
    if [[ -z $zip_file_dir ]]; then 
        zip_file_dir=".";
    fi

    if [[ ! -d $zip_file_dir ]]; then 
        terminate "The first argument '$zip_file_dir' is not a directory";
    fi

    # Discover all available zip files that match current RPi naming scheme, then process them one by one.  
    local zip_files_full_path_list=$(find_rpi_os_file_list  "$zip_file_dir" ".zip");
    
    if [ -z "$zip_files_full_path_list" ]
    then
        terminate "No RPi OS files have been discovered in the directory '$zip_file_dir'";
    else {
        
        while IFS= read -r zip_file_full_path; do {
            echo ""
            echo "================================================================================"
            echo "Processing '$zip_file_full_path' ..."
            echo "===================================="
            echo ""
            
            check_if_zip_file_exists_and_exit_if_not "$zip_file_full_path";
            
            echo "Creating tmp directories";
            local tmp_image_dir=$(mktemp -d -t rpi-image---XXXXXXXXXXXX);
            local tmp_mount_root_dir=$(mktemp -d -t rpi-mount-root---XXXXXXXXXXXX);
            
            unzip_image_to_tmp_image_dir "$zip_file_full_path" "$tmp_image_dir"
            
            
            # Discover all available img files in the tmp_image_dir, then process them one by one.  
            local img_files_full_path_list=$(find_rpi_os_file_list  "$tmp_image_dir" ".img");
            
            if [ -z "$img_files_full_path_list" ]
            then
                terminate "No RPi OS image files have been discovered in the directory '$tmp_image_dir'";
            else {


                while IFS= read -r img_file_full_path; do {
                    echo ""
                    echo "================================================================================"
                    echo "Processing image file '$img_file_full_path' ..."
                    echo "===================================="
                    echo ""
                    
                    check_if_zip_file_exists_and_exit_if_not "$img_file_full_path";

                    if mount_original_image_and_prepare_root_for_read_only_mode "$img_file_full_path" "$tmp_mount_root_dir"; then { 
                        umount "$tmp_mount_root_dir"
                        echo "Unmount modified image Root Partition."
                    }
                    else
                        echo "Mount FAILED for original Root Partition. ErrorCode=$?."
                    fi
                    
                    echo "Moving modified system image to '$modified_image_destination_directory'. The image ready for burning onto SD card."
                    move_modified_system_image "$img_file_full_path" "$modified_image_destination_directory"
                    
                }
                done <<< "$img_files_full_path_list"

            }
            fi
            
            echo "Removing tmp directories";
            rm -rf "$tmp_image_dir";
            rm -rf "$tmp_mount_root_dir";
        }   
        done <<< "$zip_files_full_path_list"
    }
    fi

    #
    #################################################################################

    echo "";
    echo "'$this_module_name' FINISHED at $(date)";
}

#################################################################################
#################################################################################
#################################################################################

main "$1";

#################################################################################
#################################################################################
#################################################################################

#echo "---------------- custom_additions --------------------";
include_custom_additions;
####################################################################
# NOTE.
# More "custom-additions-*" to come. 
# Please visit https://altomaxtech.com/ often to check out the news.
####################################################################
# If you want to automatically include a "custom-additions-*" script,
# just put it into the same directory with this script file.
####################################################################

#################################################################################
#################################################################################
#################################################################################


